"""
This module implements the handling (CRUD) of DSP users.

CREATE:
    * Instantiate a new object of the class User with all required parameters
    * Call the ``create``-method on the instance to create the new user

READ:
    * Instantiate a new objects with ``iri`` given
    * Call the ``read``-method on the instance
    * Access the information that has been ptovided to the instance

In addition there is a static methods ``getAllProjects`` which returns a list of all projects
"""

from __future__ import annotations

import urllib.parse
from typing import Any
from typing import Optional
from typing import Union
from urllib.parse import quote_plus

from dsp_tools.clients.connection import Connection
from dsp_tools.commands.project.legacy_models.group import Group
from dsp_tools.commands.project.legacy_models.model import Model
from dsp_tools.error.exceptions import BaseError
from dsp_tools.legacy_models.langstring import Languages


class User(Model):
    """
    This class represents a user in DSP.

    Attributes
    ----------

    iri : str
        IRI of the user [readonly, cannot be modified after creation of instance]

    username : str
        Unique identifier (not an IRI) of the user [read/write]

    email : str
        Email of the user [read/write]

    givenName : str
        Given name (firstname) of the user [read/write]

    familyName : str
        Family name of user (lastname) [read/write]

    password : str
        Password of user [write only]

    lang : Language
        Preferred language of the user. For setting can be Language instance or string "EN", "DE", "FR", "IT"

    status : bool
        Status of the user, If active, is set to True, otherwise false [read/write]

    in_groups : set[str]
        Set of group IRI's the user is member of [readonly].

    in_projects : set[str]
        Set of project IRI's the user belongs to
        Use ``addToproject``and ``rmFromproject`` to modify project membership


    Methods
    -------

    create : DSP user information object
        Creates a new user and returns the information about this user as it is in DSP

    read : DSP user information object
        Read user data

    getAllUsers : list of user
        Get a list of all users

    print : None
        Prints the user information to stdout


    """

    ROUTE: str = "/admin/users"
    IRI: str = ROUTE + "/iri/"
    PROJECT_MEMBERSHIPS: str = "/project-memberships/"
    PROJECT_ADMIN_MEMBERSHIPS: str = "/project-admin-memberships/"
    GROUP_MEMBERSHIPS: str = "/group-memberships/"

    _iri: str
    _username: str
    _email: str
    _givenName: str
    _familyName: str
    _password: str
    _lang: Languages
    _status: bool
    _in_groups: set[str]
    _in_projects: dict[str, bool]
    _add_to_project: dict[str, bool]
    _rm_from_project: dict[str, bool]
    _add_to_group: set[str]
    _rm_from_group: set[str]
    _change_admin: dict[str, bool]

    def __init__(
        self,
        con: Connection,
        iri: Optional[str] = None,
        username: Optional[str] = None,
        email: Optional[str] = None,
        givenName: Optional[str] = None,
        familyName: Optional[str] = None,
        password: Optional[str] = None,
        lang: Optional[Union[str, Languages]] = None,
        status: Optional[bool] = None,
        in_projects: Optional[dict[str, bool]] = None,
        in_groups: Optional[set[str]] = None,
    ):
        """
        Constructor for User

        The constructor is user internally or externally, when a new user should be created in DSP.

        :param con: Connection instance [required]
        :param iri: IRI of the user [required for CREATE, READ]
        :param username: Username [required for CREATE]
        :param email: Email address [required for CREATE]
        :param givenName: Given name (firstname) of user [required for CREATE]
        :param familyName: Family name (lastname) of user [required for CREATE]
        :param password: Password [required for CREATE]
        :param lang: Preferred language of the user [optional]
        :param status: Status (active = True, inactive/deleted = False) [optional]
        :param in_projects: Dict with project-IRI as key, boolean(True=project admin) as value [optional]
        :param in_groups: Set with group-IRI's the user should belong to [optional]
        """
        super().__init__(con)
        self._iri = iri
        self._username = str(username) if username is not None else None
        self._email = str(email) if email is not None else None
        self._givenName = str(givenName) if givenName is not None else None
        self._familyName = str(familyName) if familyName is not None else None
        self._password = str(password) if password is not None else None
        if lang is not None:
            if isinstance(lang, Languages):
                self._lang = lang
            else:
                lmap = {a.value: a for a in Languages}
                if lmap.get(lang) is None:
                    raise BaseError('Invalid language string "' + lang + '"!')
                self._lang = lmap[lang]
        else:
            self._lang = None
        self._status = None if status is None else bool(status)

        if in_projects is None or isinstance(in_projects, dict):
            self._in_projects = in_projects if in_projects is not None else {}
        else:
            raise BaseError("In_project must be tuple (project-iri: str, is-admin: bool)!")

        if in_groups is None or isinstance(in_groups, set):
            self._in_groups = in_groups if in_groups is not None else set()
        else:
            raise BaseError("In_groups must be a set of strings or None!")

    @property
    def iri(self) -> Optional[str]:
        """IRI of this user"""
        return self._iri

    @property
    def username(self) -> Optional[str]:
        """Username of this user"""
        return self._username

    @username.setter
    def username(self, value: Optional[str]) -> None:
        if value is None:
            return
        self._username = str(value)
        self._changed.add("username")

    @property
    def email(self) -> Optional[str]:
        """Email address of this user"""
        return self._email

    @email.setter
    def email(self, value: Optional[str]) -> None:
        if value is None:
            return
        self._email = str(value)
        self._changed.add("email")

    @property
    def password(self) -> Optional[str]:
        """Password of this user"""
        return self._password

    @password.setter
    def password(self, value: Optional[str]) -> None:
        if value is None:
            return
        self._password = value
        self._changed.add("password")

    @property
    def givenName(self) -> Optional[str]:
        """Given name (firstname) of this user"""
        return self._givenName

    @givenName.setter
    def givenName(self, value: Optional[str]) -> None:
        if value is None:
            return
        self._givenName = str(value)
        self._changed.add("givenName")

    @property
    def familyName(self) -> Optional[str]:
        """Family name (lastname) of this user"""
        return self._familyName

    @familyName.setter
    def familyName(self, value: Optional[str]) -> None:
        if value is None:
            return
        self._familyName = str(value)
        self._changed.add("familyName")

    @property
    def lang(self) -> Optional[Languages]:
        """Language of this user"""
        return self._lang

    @lang.setter
    def lang(self, value: Optional[Union[str, Languages]]) -> None:
        if value is None:
            return
        if isinstance(value, Languages):
            self._lang = value
            self._changed.add("lang")
        else:
            lmap = {a.value: a for a in Languages}
            if lmap.get(value) is None:
                raise BaseError('Invalid language string "' + value + '"!')
            self._lang = lmap[value]
            self._changed.add("lang")

    @property
    def status(self) -> bool:
        """Status of this user (True=active, False=inactive)"""
        return self._status

    @status.setter
    def status(self, value: Optional[bool]) -> None:
        self._status = None if value is None else bool(value)
        if value is not None:
            self._changed.add("status")

    @property
    def in_groups(self) -> set[str]:
        """Set of group IRI's the user is member of"""
        return self._in_groups

    @property
    def in_projects(self) -> dict[str, bool]:
        """dict with project-IRI as key, boolean(True=project admin) as value"""
        return self._in_projects

    @classmethod
    def _make_fromJsonObj(cls, con: Connection, json_obj: dict[str, Any]) -> User:
        User._check_if_jsonObj_has_required_info(json_obj)

        in_groups, in_projects = cls._update_permissions_and_groups(json_obj)

        return cls(
            con=con,
            iri=json_obj["id"],
            username=json_obj["username"],
            email=json_obj["email"],
            givenName=json_obj.get("givenName"),
            familyName=json_obj.get("familyName"),
            lang=json_obj.get("lang"),
            status=json_obj["status"],
            in_projects=in_projects,
            in_groups=in_groups,
        )

    @staticmethod
    def _check_if_jsonObj_has_required_info(json_obj: dict[str, Any]) -> None:
        problems = []
        if json_obj.get("id") is None:
            problems.append('User "iri" is missing in JSON from DSP')

        if json_obj.get("email") is None:
            problems.append('User "email" is missing in JSON from DSP')

        if json_obj.get("username") is None:
            problems.append('User "email" is missing in JSON from DSP')

        if json_obj.get("status") is None:
            problems.append('User "email" is missing in JSON from DSP')

        if problems:
            raise BaseError("\n".join(problems))

    @classmethod
    def _update_permissions_and_groups(cls, json_obj: dict[str, Any]) -> tuple[set[str], dict[str, bool]]:
        in_projects: dict[str, bool] = {}
        in_groups: set[str] = set()
        if json_obj.get("permissions") is not None and json_obj["permissions"].get("groupsPerProject") is not None:
            for project_iri, group_memberships in json_obj["permissions"]["groupsPerProject"].items():
                for group in group_memberships:
                    if group == Group.PROJECT_MEMBER_GROUP:
                        if in_projects.get(project_iri) is None:
                            in_projects[project_iri] = False
                    elif group == Group.PROJECT_ADMIN_GROUP:
                        in_projects[project_iri] = True
                    else:
                        in_groups.add(group)
            return in_groups, in_projects
        return in_groups, in_projects

    def create(self) -> User:
        """
        Create new user in DSP

        :return: JSON-object from DSP
        """
        jsonobj = self._toJosonObj_create()
        result = self._con.post(User.ROUTE, jsonobj)
        iri = result["user"]["id"]
        if self._in_projects is not None:
            for project in self._in_projects:
                result = self._con.post(User.IRI + quote_plus(iri) + User.PROJECT_MEMBERSHIPS + quote_plus(project))
                if self._in_projects[project]:
                    result = self._con.post(
                        User.IRI + quote_plus(iri) + User.PROJECT_ADMIN_MEMBERSHIPS + quote_plus(project)
                    )
        if self._in_groups is not None:
            for group in self._in_groups:
                result = self._con.post(User.IRI + quote_plus(iri) + User.GROUP_MEMBERSHIPS + quote_plus(group))
        return User._make_fromJsonObj(self._con, result["user"])

    def _toJosonObj_create(self):
        tmp = {}
        if self._username is None:
            raise BaseError("There must be a valid username!")
        tmp["username"] = self._username
        if self._email is None:
            raise BaseError("'email' is mandatory!")
        tmp["email"] = self._email
        if self._givenName is None:
            raise BaseError("'givenName is mandatory!")
        tmp["givenName"] = self._givenName
        if self._familyName is None:
            raise BaseError("'familyName' is mandatory!")
        tmp["familyName"] = self._familyName
        if self._password is None:
            raise BaseError("'password' is mandatory!")
        tmp["password"] = self._password
        if self._lang is None:
            raise BaseError("'language' is mandatory!")
        tmp["lang"] = self._lang.value
        tmp["status"] = True if self._status is None else self._status
        # the API expects the key "systemAdmin" to be present. DSP-TOOLS doesn't support creating SystemAdmins any more.
        tmp["systemAdmin"] = False
        return tmp

    def read(self) -> User:
        """
        Read the user information from DSP. The User object must have a valid iri or email!

        :return: JSON-object from DSP
        """
        if self._iri is not None:
            result = self._con.get(User.IRI + quote_plus(self._iri))
        elif self._email is not None:
            result = self._con.get(User.ROUTE + "/email/" + quote_plus(self._email))
        elif self._username is not None:
            result = self._con.get(User.ROUTE + "/username/" + quote_plus(self._username))
        else:
            raise BaseError("Either user-iri or email is required!")
        return User._make_fromJsonObj(self._con, result["user"])

    @staticmethod
    def getAllUsers(con: Connection) -> list[User]:
        """
        Get a list of all users (static method)

        :param con: Connection instance
        :return: List of users
        """

        result = con.get(User.ROUTE)
        if "users" not in result:
            raise BaseError("Request got no users!")
        return [User._make_fromJsonObj(con, a) for a in result["users"]]

    @staticmethod
    def getAllUsersForProject(con: Connection, proj_shortcode: str) -> Optional[list[User]]:
        """
        Get a list of all users that belong to a project (static method)

        :param con: Connection instance
        :project_shortcode: Shortcode of the project
        :return: List of users belonging to that project
        """
        members = con.get(f"/admin/projects/shortcode/{proj_shortcode}/members")
        if members is None or len(members) < 1:
            return None
        res: list[User] = [User._make_fromJsonObj(con, a) for a in members["members"]]
        res.reverse()
        return res

    def createDefinitionFileObj(
        self,
        con: Connection,
        proj_shortname: str,
        proj_iri: str,
    ) -> dict[str, Union[str, list[str], bool, None]]:
        """Create a JSON object that can be used to create a project definition file"""
        user: dict[str, Union[str, list[str], bool, None]] = {
            "username": self.username,
            "email": self._email,
            "givenName": self._givenName,
            "familyName": self._familyName,
            "password": "",
        }
        if self._lang:
            user["lang"] = self._lang.value
        groups = list()
        for group_iri in self._in_groups:
            group_info = con.get(f"/admin/groups/{urllib.parse.quote_plus(group_iri)}")
            if "group" in group_info and "name" in group_info["group"]:
                groupname = group_info["group"]["name"]
                groups.append(f"{proj_shortname}:{groupname}")
        user["groups"] = groups
        user["projects"] = list()
        for proj, is_admin in self._in_projects.items():
            if proj == proj_iri:
                if is_admin:
                    user["projects"].append(f"{proj_shortname}:admin")
                else:
                    user["projects"].append(f"{proj_shortname}:member")
        user["status"] = self._status
        return user



from pathlib import Path

from dsp_tools.clients.authentication_client import AuthenticationClient
from dsp_tools.commands.validate_data.models.validation import RDFGraphs
from dsp_tools.commands.validate_data.prepare_data.prepare_data import get_info_and_parsed_resources_from_file
from dsp_tools.commands.validate_data.prepare_data.prepare_data import prepare_data_for_validation_from_parsed_resource


def prepare_data_for_validation_from_file(
    filepath: Path, auth: AuthenticationClient, ignore_duplicate_files_warning: bool
) -> tuple[RDFGraphs, set[str]]:
    parsed_resources, shortcode, authorship_lookup, permission_ids = get_info_and_parsed_resources_from_file(
        file=filepath,
        api_url=auth.server,
    )
    return prepare_data_for_validation_from_parsed_resource(
        parsed_resources=parsed_resources,
        authorship_lookup=authorship_lookup,
        permission_ids=permission_ids,
        auth=auth,
        shortcode=shortcode,
        ignore_duplicate_files_warning=ignore_duplicate_files_warning,
    )



from datetime import datetime
from pathlib import Path
from typing import cast

import pandas as pd
from loguru import logger

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.cli.args import ValidateDataConfig
from dsp_tools.cli.args import ValidationSeverity
from dsp_tools.clients.authentication_client import AuthenticationClient
from dsp_tools.clients.authentication_client_live import AuthenticationClientLive
from dsp_tools.commands.validate_data.models.input_problems import OntologyValidationProblem
from dsp_tools.commands.validate_data.models.input_problems import SortedProblems
from dsp_tools.commands.validate_data.models.input_problems import UnknownClassesInData
from dsp_tools.commands.validate_data.models.input_problems import ValidateDataResult
from dsp_tools.commands.validate_data.models.validation import RDFGraphs
from dsp_tools.commands.validate_data.models.validation import ValidationReportGraphs
from dsp_tools.commands.validate_data.prepare_data.prepare_data import get_info_and_parsed_resources_from_file
from dsp_tools.commands.validate_data.prepare_data.prepare_data import prepare_data_for_validation_from_parsed_resource
from dsp_tools.commands.validate_data.process_validation_report.get_user_validation_message import get_user_message
from dsp_tools.commands.validate_data.process_validation_report.get_user_validation_message import sort_user_problems
from dsp_tools.commands.validate_data.process_validation_report.query_validation_result import reformat_validation_graph
from dsp_tools.commands.validate_data.shacl_cli_validator import ShaclCliValidator
from dsp_tools.commands.validate_data.validation.check_duplicate_files import check_for_duplicate_files
from dsp_tools.commands.validate_data.validation.check_for_unknown_classes import check_for_unknown_resource_classes
from dsp_tools.commands.validate_data.validation.check_for_unknown_classes import get_msg_str_unknown_classes_in_data
from dsp_tools.commands.validate_data.validation.get_validation_report import get_validation_report
from dsp_tools.commands.validate_data.validation.validate_ontology import get_msg_str_ontology_validation_violation
from dsp_tools.commands.validate_data.validation.validate_ontology import validate_ontology
from dsp_tools.error.exceptions import BaseError
from dsp_tools.utils.ansi_colors import BACKGROUND_BOLD_CYAN
from dsp_tools.utils.ansi_colors import BACKGROUND_BOLD_GREEN
from dsp_tools.utils.ansi_colors import BACKGROUND_BOLD_RED
from dsp_tools.utils.ansi_colors import BACKGROUND_BOLD_YELLOW
from dsp_tools.utils.ansi_colors import BOLD_CYAN
from dsp_tools.utils.ansi_colors import BOLD_RED
from dsp_tools.utils.ansi_colors import BOLD_YELLOW
from dsp_tools.utils.ansi_colors import RESET_TO_DEFAULT
from dsp_tools.utils.data_formats.uri_util import is_prod_like_server
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedResource

VALIDATION_ERRORS_FOUND_MSG = BACKGROUND_BOLD_RED + "\n   Validation errors found!   " + RESET_TO_DEFAULT
NO_VALIDATION_ERRORS_FOUND_MSG = BACKGROUND_BOLD_GREEN + "\n   No validation errors found!   " + RESET_TO_DEFAULT


def validate_data(
    filepath: Path, creds: ServerCredentials, ignore_duplicate_files_warning: bool, save_graphs: bool
) -> bool:
    """
    Takes a file and project information and validates it against the ontologies on the server.

    Args:
        filepath: path to the xml data file
        creds: server credentials for authentication
        ignore_duplicate_files_warning: ignore the shape that checks for duplicate files
        save_graphs: if this flag is set, all the graphs will be saved in a folder

    Returns:
        True if no errors that impede an xmlupload were found.
        Warnings and user info do not impede an xmlupload.
    """
    graph_save_dir = None

    if save_graphs:
        graph_save_dir = _get_graph_save_dir(filepath)
    config = ValidateDataConfig(
        xml_file=filepath,
        save_graph_dir=graph_save_dir,
        severity=ValidationSeverity.INFO,
        ignore_duplicate_files_warning=ignore_duplicate_files_warning,
        is_on_prod_server=is_prod_like_server(creds.server),
    )
    auth = AuthenticationClientLive(server=creds.server, email=creds.user, password=creds.password)

    parsed_resources, shortcode, authorship_lookup, permission_ids = get_info_and_parsed_resources_from_file(
        file=filepath,
        api_url=auth.server,
    )
    return validate_parsed_resources(
        parsed_resources=parsed_resources,
        authorship_lookup=authorship_lookup,
        permission_ids=permission_ids,
        shortcode=shortcode,
        config=config,
        auth=auth,
    )


def validate_parsed_resources(
    parsed_resources: list[ParsedResource],
    authorship_lookup: dict[str, list[str]],
    permission_ids: list[str],
    shortcode: str,
    config: ValidateDataConfig,
    auth: AuthenticationClient,
) -> bool:
    ignore_duplicate_files = config.ignore_duplicate_files_warning
    if not config.ignore_duplicate_files_warning:
        duplicate_check = check_for_duplicate_files(parsed_resources, config)
        if duplicate_check.user_msg:
            logger.error(duplicate_check.user_msg)
            print(f"{BACKGROUND_BOLD_RED}   Duplicate Filepaths Found!    {RESET_TO_DEFAULT}")
            print(f"{BOLD_RED}{duplicate_check.user_msg}{RESET_TO_DEFAULT}")
        if not duplicate_check.should_continue:
            return False
        ignore_duplicate_files = duplicate_check.duplicate_files_must_be_ignored

    rdf_graphs, used_iris = prepare_data_for_validation_from_parsed_resource(
        parsed_resources=parsed_resources,
        authorship_lookup=authorship_lookup,
        permission_ids=permission_ids,
        auth=auth,
        shortcode=shortcode,
        ignore_duplicate_files_warning=ignore_duplicate_files,
    )
    validation_result = _validate_data(rdf_graphs, used_iris, config)
    if validation_result.no_problems:
        logger.debug("No validation errors found.")
        print(NO_VALIDATION_ERRORS_FOUND_MSG)
        return True
    if isinstance(validation_result.problems, UnknownClassesInData):
        msg = get_msg_str_unknown_classes_in_data(validation_result.problems)
        logger.error(msg)
        print(VALIDATION_ERRORS_FOUND_MSG)
        print(msg)
        # if unknown classes are found, we cannot validate all the data in the file
        return False
    if isinstance(validation_result.problems, OntologyValidationProblem):
        msg = get_msg_str_ontology_validation_violation(validation_result.problems)
        logger.error(msg)
        print(VALIDATION_ERRORS_FOUND_MSG)
        print(msg)
        # if the ontology itself has errors, we will not validate the data
        return False
    if isinstance(validation_result.problems, SortedProblems):
        _print_shacl_validation_violation_message(validation_result.problems, validation_result.report_graphs, config)
        return _get_validation_status(validation_result.problems, config.is_on_prod_server)
    else:
        raise BaseError(f"Unknown validate data problems: {validation_result.problems!s}")


def _validate_data(graphs: RDFGraphs, used_iris: set[str], config: ValidateDataConfig) -> ValidateDataResult:
    logger.debug(f"Validate-data called with the following config: {vars(config)}")
    if unknown_classes := check_for_unknown_resource_classes(graphs, used_iris):
        return ValidateDataResult(False, unknown_classes, None)
    shacl_validator = ShaclCliValidator()
    onto_validation_result = validate_ontology(graphs.ontos, shacl_validator, config)
    if onto_validation_result:
        return ValidateDataResult(False, onto_validation_result, None)
    report = get_validation_report(graphs, shacl_validator, config.save_graph_dir)
    if report.conforms:
        return ValidateDataResult(True, None, None)
    reformatted = reformat_validation_graph(report)
    sorted_problems = sort_user_problems(reformatted)
    return ValidateDataResult(False, sorted_problems, report)


def _get_graph_save_dir(filepath: Path) -> Path:
    parent_directory = filepath.parent
    new_directory = parent_directory / "graphs"
    new_directory.mkdir(exist_ok=True)
    save_file_template = new_directory / filepath.stem
    print(BOLD_CYAN + f"\n   Saving graphs to {save_file_template}   " + RESET_TO_DEFAULT)
    return save_file_template


def _get_validation_status(all_problems: SortedProblems, is_on_prod: bool) -> bool:
    violations = any(
        [
            bool(all_problems.unique_violations),
            bool(all_problems.unexpected_shacl_validation_components),
        ]
    )
    if violations:
        return False
    if is_on_prod and all_problems.user_warnings:
        return False
    return True


def _print_shacl_validation_violation_message(
    sorted_problems: SortedProblems, report: ValidationReportGraphs | None, config: ValidateDataConfig
) -> None:
    messages = get_user_message(sorted_problems, config.severity)
    if messages.violations:
        print(VALIDATION_ERRORS_FOUND_MSG)
        print(BOLD_RED, messages.violations.message_header, RESET_TO_DEFAULT)
        v_body = messages.violations.message_body
        if messages.violations.message_df is not None:
            v_body = _save_message_df_get_message_body(messages.violations.message_df, "error", config.xml_file)
        print(v_body)
        logger.error(messages.violations.message_header, v_body)
    else:
        logger.debug("No validation errors found.")
        print(NO_VALIDATION_ERRORS_FOUND_MSG)
    if messages.warnings and config.severity.value <= 2:
        print(BACKGROUND_BOLD_YELLOW + "\n    Warning!    " + RESET_TO_DEFAULT)
        print(BOLD_YELLOW, messages.warnings.message_header, RESET_TO_DEFAULT)
        w_body = messages.warnings.message_body
        if messages.warnings.message_df is not None:
            w_body = _save_message_df_get_message_body(messages.warnings.message_df, "warning", config.xml_file)
        print(w_body)
        logger.warning(messages.warnings.message_header, w_body)
    if messages.infos and config.severity.value == 1:
        print(BACKGROUND_BOLD_CYAN + "\n    Potential Problems Found    " + RESET_TO_DEFAULT)
        print(BOLD_CYAN, messages.infos.message_header, RESET_TO_DEFAULT)
        i_body = messages.infos.message_body
        if messages.infos.message_df is not None:
            i_body = _save_message_df_get_message_body(messages.infos.message_df, "info", config.xml_file)
        print(i_body)
        logger.info(messages.infos.message_header, i_body)
    if messages.unexpected_violations:
        logger.error(messages.unexpected_violations.message_header, messages.unexpected_violations.message_body)
        print(
            BACKGROUND_BOLD_RED,
            "\n    Unknown violations found!   ",
            RESET_TO_DEFAULT,
        )
        if config.save_graph_dir:
            print(
                BOLD_RED,
                messages.unexpected_violations.message_header,
                "Consult the saved graphs for details.",
                RESET_TO_DEFAULT,
            )
            print(messages.unexpected_violations.message_body)
        else:
            report_graph = cast(ValidationReportGraphs, report)
            _save_unexpected_results_and_inform_user(report_graph, config.xml_file)


def _save_message_df_get_message_body(df: pd.DataFrame, severity: str, file_path: Path) -> str:
    out_path = file_path.parent / f"{file_path.stem}_validation_{severity}.csv"
    msg = f"Due to the large number of violations the information was saved at '{out_path}'"
    df.to_csv(out_path, index=False)
    return msg


def _save_unexpected_results_and_inform_user(report: ValidationReportGraphs, filepath: Path) -> None:
    timestamp = f"{datetime.now()!s}_"
    save_path = filepath.parent / f"{timestamp}_validation_result.ttl"
    report.validation_graph.serialize(save_path)
    shacl_p = filepath.parent / f"{timestamp}_shacl.ttl"
    report.shacl_graph.serialize(shacl_p)
    data_p = filepath.parent / f"{timestamp}_data.ttl"
    report.data_graph.serialize(data_p)
    msg = (
        f"\nPlease contact the development team with the files starting with the timestamp '{timestamp}' "
        f"in the directory '{filepath.parent}'."
    )
    print(BOLD_RED + msg + RESET_TO_DEFAULT)



# mypy: disable-error-code="method-assign,no-untyped-def"

import warnings

import pytest
import regex

from dsp_tools.error.xmllib_warnings import XmllibInputWarning
from dsp_tools.xmllib import VideoSegmentResource
from dsp_tools.xmllib.models.config_options import Permissions
from dsp_tools.xmllib.models.dsp_base_resources import AudioSegmentResource
from dsp_tools.xmllib.models.dsp_base_resources import LinkResource
from dsp_tools.xmllib.models.dsp_base_resources import RegionResource
from dsp_tools.xmllib.models.dsp_base_resources import SegmentBounds
from dsp_tools.xmllib.models.dsp_base_resources import _check_strings
from dsp_tools.xmllib.models.dsp_base_resources import _warn_value_exists


class TestRegionResource:
    def test_good(self):
        with warnings.catch_warnings(record=True) as caught_warnings:
            RegionResource.create_new("id", "lbl", "regionOfId")
        assert len(caught_warnings) == 0

    def test_warns(self):
        with warnings.catch_warnings(record=True) as caught_warnings:
            RegionResource.create_new("", "", "")
        assert len(caught_warnings) == 3


class TestLinkResource:
    def test_good(self):
        with warnings.catch_warnings(record=True) as caught_warnings:
            LinkResource.create_new("id", "lbl", ["link_to"])
        assert len(caught_warnings) == 0

    def test_warns(self):
        with warnings.catch_warnings(record=True) as caught_warnings:
            LinkResource.create_new("", "", [])
        assert len(caught_warnings) == 3


def test_segment_bounds():
    expected = regex.escape(
        " | Resource ID 'res_id' | Segment bounds must be a float or integer. "
        "The following places have an unexpected type: \n"
        "    - Segment Start Value: string | Type: <class 'str'>"
    )
    with pytest.warns(XmllibInputWarning, match=expected):
        SegmentBounds("string", "1", Permissions.PROJECT_SPECIFIC_PERMISSIONS, "res_id")


class TestVideoSegmentResource:
    def test_good(self):
        with warnings.catch_warnings(record=True) as caught_warnings:
            VideoSegmentResource.create_new("id", "lbl", "segment_of", "0", "1")
        assert len(caught_warnings) == 0

    def test_warns_strings(self):
        with warnings.catch_warnings(record=True) as caught_warnings:
            VideoSegmentResource.create_new("", "", "", "0", "1")
        assert len(caught_warnings) == 3

    def test_warns_segment_bounds(self):
        with warnings.catch_warnings(record=True) as caught_warnings:
            VideoSegmentResource.create_new("id", "lbl", "segment_of", "", "")
        assert len(caught_warnings) == 1


class TestAudioSegmentResource:
    def test_good(self):
        with warnings.catch_warnings(record=True) as caught_warnings:
            AudioSegmentResource.create_new("id", "lbl", "segment_of", "0", "1")
        assert len(caught_warnings) == 0

    def test_warns_strings(self):
        with warnings.catch_warnings(record=True) as caught_warnings:
            AudioSegmentResource.create_new("", "", "", "0", "1")
        assert len(caught_warnings) == 3

    def test_warns_segment_bounds(self):
        with warnings.catch_warnings(record=True) as caught_warnings:
            AudioSegmentResource.create_new("id", "lbl", "segment_of", "", "")
        assert len(caught_warnings) == 1


# The warning matches are constructed so that the line number is not checked as that is prone to change.


def test_check_strings_prop():
    expected = (
        regex.escape("File 'test_dsp_base_resources.py:")
        + r"\d+"
        + regex.escape("' | Resource ID 'id' | Property 'prop' | The entered string is not valid.")
    )
    with pytest.warns(XmllibInputWarning, match=expected):
        _check_strings(string_to_check="", res_id="id", prop_name="prop")


def test_check_strings_field():
    expected = (
        regex.escape("File 'test_dsp_base_resources.py:")
        + r"\d+"
        + regex.escape("' | Resource ID 'id' | Field 'field' | The entered string is not valid.")
    )
    with pytest.warns(XmllibInputWarning, match=expected):
        _check_strings(string_to_check="", res_id="id", field_name="field")


def test_warn_value_exists():
    expected = (
        regex.escape("File 'test_dsp_base_resources.py:")
        + r"\d+"
        + regex.escape(
            "' | Resource ID 'id' | Field 'field' | "
            "This resource already has a value in this location. The old value 'old' is being replace with 'new'."
        )
    )
    with pytest.warns(XmllibInputWarning, match=expected):
        _warn_value_exists(old_value="old", new_value="new", res_id="id", value_field="field")



from pathlib import Path
from unittest.mock import Mock
from unittest.mock import patch

import pytest

from dsp_tools.cli import entry_point
from dsp_tools.cli.args import ServerCredentials
from dsp_tools.cli.args import ValidationSeverity
from dsp_tools.commands.start_stack import StackConfiguration
from dsp_tools.commands.xmlupload.upload_config import UploadConfig

EXIT_CODE_TWO = 2


def test_invalid_arguments() -> None:
    args = "invalid".split()
    with pytest.raises(SystemExit) as ex:
        entry_point.run(args)
    assert ex.value.code == EXIT_CODE_TWO


@patch("dsp_tools.cli.call_action.validate_lists_section_with_schema")
def test_lists_validate(validate_lists: Mock) -> None:
    file = "filename.json"
    args = f"create --lists-only --validate-only {file}".split()
    entry_point.run(args)
    validate_lists.assert_called_once_with(file)


@patch("dsp_tools.cli.call_action.create_only_lists")
def test_lists_create(create_lists: Mock) -> None:
    create_lists.return_value = ({}, True)
    file = "filename.json"
    args = f"create --lists-only {file}".split()
    creds = ServerCredentials(server="http://0.0.0.0:3333", user="root@example.com", password="test")
    entry_point.run(args)
    create_lists.assert_called_once_with(
        project_file_as_path_or_parsed=file,
        creds=creds,
    )


@patch("dsp_tools.cli.call_action.validate_project")
def test_project_validate(validate_project: Mock) -> None:
    file = "filename.json"
    args = f"create --validate-only {file}".split()
    entry_point.run(args)
    validate_project.assert_called_once_with(file)


@patch("dsp_tools.cli.call_action.create_project")
def test_project_create(create_project: Mock) -> None:
    file = "filename.json"
    args = f"create {file}".split()
    creds = ServerCredentials(server="http://0.0.0.0:3333", user="root@example.com", password="test")
    entry_point.run(args)
    create_project.assert_called_once_with(
        project_file_as_path_or_parsed=file,
        creds=creds,
        verbose=False,
    )


@patch("dsp_tools.cli.call_action.get_project")
def test_project_get(get_project: Mock) -> None:
    file = "filename.json"
    project = "shortname"
    args = f"get --project {project} {file}".split()
    creds = ServerCredentials(server="http://0.0.0.0:3333", user="root@example.com", password="test")
    entry_point.run(args)
    get_project.assert_called_once_with(
        project_identifier=project,
        outfile_path=file,
        creds=creds,
        verbose=False,
    )


@patch("dsp_tools.cli.call_action.xmlupload")
def test_xmlupload_default(xmlupload: Mock) -> None:
    file = "filename.xml"
    args = f"xmlupload {file}".split()
    creds = ServerCredentials(
        server="http://0.0.0.0:3333",
        user="root@example.com",
        password="test",
        dsp_ingest_url="http://0.0.0.0:3340",
    )
    entry_point.run(args)
    xmlupload.assert_called_once_with(
        input_file=Path(file),
        creds=creds,
        imgdir=".",
        config=UploadConfig(
            skip_iiif_validation=False,
            interrupt_after=None,
            validation_severity=ValidationSeverity.INFO,
        ),
    )


@patch("dsp_tools.cli.call_action.parse_and_validate_xml_file")
def test_xmlupload_validate(validate_xml: Mock) -> None:
    file = "filename.xml"
    args = f"xmlupload --validate-only {file}".split()
    entry_point.run(args)
    validate_xml.assert_called_once_with(Path(file))


@patch("dsp_tools.cli.call_action.xmlupload")
def test_xmlupload_no_iiif(xmlupload: Mock) -> None:
    file = "filename.xml"
    no_validation = "--no-iiif-uri-validation"
    args = f"xmlupload {no_validation} {file}".split()
    creds = ServerCredentials(
        server="http://0.0.0.0:3333",
        user="root@example.com",
        password="test",
        dsp_ingest_url="http://0.0.0.0:3340",
    )
    entry_point.run(args)
    xmlupload.assert_called_once_with(
        input_file=Path(file),
        creds=creds,
        imgdir=".",
        config=UploadConfig(skip_iiif_validation=True),
    )


@patch("dsp_tools.cli.call_action.xmlupload")
def test_xmlupload_skip_validation(xmlupload: Mock) -> None:
    file = "filename.xml"
    no_validation = "--skip-validation"
    args = f"xmlupload {no_validation} {file}".split()
    creds = ServerCredentials(
        server="http://0.0.0.0:3333",
        user="root@example.com",
        password="test",
        dsp_ingest_url="http://0.0.0.0:3340",
    )
    entry_point.run(args)
    xmlupload.assert_called_once_with(
        input_file=Path(file),
        creds=creds,
        imgdir=".",
        config=UploadConfig(skip_validation=True),
    )


@patch("dsp_tools.cli.call_action.xmlupload")
def test_xmlupload_ignore_duplicate_files_warning(xmlupload: Mock) -> None:
    file = "filename.xml"
    ignore_duplicate_files = "--ignore-duplicate-files-warning"
    args = f"xmlupload {ignore_duplicate_files} {file}".split()
    creds = ServerCredentials(
        server="http://0.0.0.0:3333",
        user="root@example.com",
        password="test",
        dsp_ingest_url="http://0.0.0.0:3340",
    )
    entry_point.run(args)
    xmlupload.assert_called_once_with(
        input_file=Path(file),
        creds=creds,
        imgdir=".",
        config=UploadConfig(ignore_duplicate_files_warning=True),
    )


@patch("dsp_tools.cli.call_action.xmlupload")
def test_xmlupload_default_validation_severity_warning(xmlupload: Mock) -> None:
    file = "filename.xml"
    args = f"xmlupload {file} --validation-severity warning".split()
    creds = ServerCredentials(
        server="http://0.0.0.0:3333",
        user="root@example.com",
        password="test",
        dsp_ingest_url="http://0.0.0.0:3340",
    )
    entry_point.run(args)
    xmlupload.assert_called_once_with(
        input_file=Path(file),
        creds=creds,
        imgdir=".",
        config=UploadConfig(
            skip_iiif_validation=False,
            interrupt_after=None,
            validation_severity=ValidationSeverity.WARNING,
        ),
    )


@patch("dsp_tools.cli.call_action.xmlupload")
def test_xmlupload_default_validation_severity_error(xmlupload: Mock) -> None:
    file = "filename.xml"
    args = f"xmlupload {file} --validation-severity error".split()
    creds = ServerCredentials(
        server="http://0.0.0.0:3333",
        user="root@example.com",
        password="test",
        dsp_ingest_url="http://0.0.0.0:3340",
    )
    entry_point.run(args)
    xmlupload.assert_called_once_with(
        input_file=Path(file),
        creds=creds,
        imgdir=".",
        config=UploadConfig(
            skip_iiif_validation=False,
            interrupt_after=None,
            validation_severity=ValidationSeverity.ERROR,
        ),
    )


@patch("dsp_tools.cli.call_action.xmlupload")
def test_xmlupload_interrupt_after(xmlupload: Mock) -> None:
    file = "filename.xml"
    args = f"xmlupload --interrupt-after=1 {file}".split()
    creds = ServerCredentials(
        server="http://0.0.0.0:3333",
        user="root@example.com",
        password="test",
        dsp_ingest_url="http://0.0.0.0:3340",
    )
    entry_point.run(args)
    xmlupload.assert_called_once_with(
        input_file=Path(file), creds=creds, imgdir=".", config=UploadConfig(interrupt_after=1)
    )


@patch("dsp_tools.cli.call_action.validate_data")
def test_validate_data_default(validate_data: Mock) -> None:
    file = "filename.xml"
    args = f"validate-data {file}".split()
    entry_point.run(args)
    creds = ServerCredentials(
        user="root@example.com", password="test", server="http://0.0.0.0:3333", dsp_ingest_url="http://0.0.0.0:3340"
    )
    validate_data.assert_called_once_with(
        filepath=Path(file), save_graphs=False, creds=creds, ignore_duplicate_files_warning=False
    )


@patch("dsp_tools.cli.call_action.validate_data")
def test_validate_data_ignore_duplicate_files(validate_data: Mock) -> None:
    file = "filename.xml"
    ignore_duplicate_files = "--ignore-duplicate-files-warning"
    args = f"validate-data {ignore_duplicate_files} {file}".split()
    entry_point.run(args)
    creds = ServerCredentials(
        user="root@example.com", password="test", server="http://0.0.0.0:3333", dsp_ingest_url="http://0.0.0.0:3340"
    )
    validate_data.assert_called_once_with(
        filepath=Path(file), save_graphs=False, creds=creds, ignore_duplicate_files_warning=True
    )


@patch("dsp_tools.cli.call_action.validate_data")
def test_validate_data_save_graph(validate_data: Mock) -> None:
    file = "filename.xml"
    args = f"validate-data {file} --save-graphs".split()
    entry_point.run(args)
    creds = ServerCredentials(
        user="root@example.com", password="test", server="http://0.0.0.0:3333", dsp_ingest_url="http://0.0.0.0:3340"
    )
    validate_data.assert_called_once_with(
        filepath=Path(file), save_graphs=True, creds=creds, ignore_duplicate_files_warning=False
    )


@patch("dsp_tools.cli.call_action.validate_data")
def test_validate_data_other_server(validate_data: Mock) -> None:
    file = "filename.xml"
    args = f"validate-data {file} -s https://api.dasch.swiss".split()
    entry_point.run(args)
    creds = ServerCredentials(
        user="root@example.com",
        password="test",
        server="https://api.dasch.swiss",
        dsp_ingest_url="https://ingest.dasch.swiss",
    )
    validate_data.assert_called_once_with(
        filepath=Path(file), save_graphs=False, creds=creds, ignore_duplicate_files_warning=False
    )


@patch("dsp_tools.cli.call_action.validate_data")
def test_validate_data_other_creds(validate_data: Mock) -> None:
    file = "filename.xml"
    server = "https://api.test.dasch.swiss"
    user = "first-name.second-name@dasch.swiss"
    password = "foobar"
    args = f"validate-data {file} --server={server} --user {user} --password={password}".split()
    entry_point.run(args)
    creds = ServerCredentials(
        user=user, password=password, server=server, dsp_ingest_url="https://ingest.test.dasch.swiss"
    )
    validate_data.assert_called_once_with(
        filepath=Path(file), save_graphs=False, creds=creds, ignore_duplicate_files_warning=False
    )


@patch("dsp_tools.cli.call_action.resume_xmlupload")
def test_resume_xmlupload_default(resume_xmlupload: Mock) -> None:
    args = "resume-xmlupload".split()
    creds = ServerCredentials(
        server="http://0.0.0.0:3333",
        user="root@example.com",
        password="test",
        dsp_ingest_url="http://0.0.0.0:3340",
    )
    entry_point.run(args)
    resume_xmlupload.assert_called_once_with(creds=creds, skip_first_resource=False)


@patch("dsp_tools.cli.call_action.resume_xmlupload")
def test_resume_xmlupload_skip_first_resource(resume_xmlupload: Mock) -> None:
    args = "resume-xmlupload --skip-first-resource".split()
    creds = ServerCredentials(
        server="http://0.0.0.0:3333",
        user="root@example.com",
        password="test",
        dsp_ingest_url="http://0.0.0.0:3340",
    )
    entry_point.run(args)
    resume_xmlupload.assert_called_once_with(creds=creds, skip_first_resource=True)


@patch("dsp_tools.cli.call_action.upload_files")
def test_upload_files_localhost(upload_files: Mock) -> None:
    file = "filename.xml"
    args = f"upload-files {file}".split()
    entry_point.run(args)
    creds = ServerCredentials(
        server="http://0.0.0.0:3333",
        user="root@example.com",
        password="test",
        dsp_ingest_url="http://0.0.0.0:3340",
    )
    upload_files.assert_called_once_with(xml_file=Path(file), creds=creds, imgdir=Path("."))


@patch("dsp_tools.cli.call_action.upload_files")
def test_upload_files_remote(upload_files: Mock) -> None:
    file = "filename.xml"
    server = "https://api.test.dasch.swiss"
    user = "first-name.second-name@dasch.swiss"
    password = "foobar"
    args = f"upload-files --server={server} --user {user} --password={password} {file}".split()
    entry_point.run(args)
    creds = ServerCredentials(
        server=server,
        user=user,
        password=password,
        dsp_ingest_url=server.replace("api", "ingest"),
    )
    upload_files.assert_called_once_with(xml_file=Path(file), creds=creds, imgdir=Path("."))


@patch("dsp_tools.cli.call_action.ingest_files")
def test_ingest_files_localhost(ingest_files: Mock) -> None:
    shortcode = "1234"
    args = f"ingest-files {shortcode}".split()
    entry_point.run(args)
    creds = ServerCredentials(
        server="http://0.0.0.0:3333",
        user="root@example.com",
        password="test",
        dsp_ingest_url="http://0.0.0.0:3340",
    )
    ingest_files.assert_called_once_with(creds=creds, shortcode=shortcode)


@patch("dsp_tools.cli.call_action.ingest_files")
def test_ingest_files_remote(ingest_files: Mock) -> None:
    shortcode = "1234"
    server = "https://api.test.dasch.swiss"
    user = "first-name.second-name@dasch.swiss"
    password = "foobar"
    args = f"ingest-files --server={server} --user {user} --password={password} {shortcode}".split()
    entry_point.run(args)
    creds = ServerCredentials(
        server=server,
        user=user,
        password=password,
        dsp_ingest_url=server.replace("api", "ingest"),
    )
    ingest_files.assert_called_once_with(creds=creds, shortcode=shortcode)


@patch("dsp_tools.cli.call_action.ingest_xmlupload")
def test_ingest_xmlupload_localhost(ingest_xmlupload: Mock) -> None:
    xml_file = Path("filename.xml")
    args = f"ingest-xmlupload {xml_file}".split()
    entry_point.run(args)
    creds = ServerCredentials(
        server="http://0.0.0.0:3333",
        user="root@example.com",
        password="test",
        dsp_ingest_url="http://0.0.0.0:3340",
    )
    ingest_xmlupload.assert_called_once_with(
        xml_file=xml_file, creds=creds, interrupt_after=None, skip_validation=False
    )


@patch("dsp_tools.cli.call_action.ingest_xmlupload")
def test_ingest_xmlupload_skip_validation(ingest_xmlupload: Mock) -> None:
    xml_file = Path("filename.xml")
    skip_validation = "--skip-validation"
    args = f"ingest-xmlupload {skip_validation} {xml_file}".split()
    entry_point.run(args)
    creds = ServerCredentials(
        server="http://0.0.0.0:3333",
        user="root@example.com",
        password="test",
        dsp_ingest_url="http://0.0.0.0:3340",
    )
    ingest_xmlupload.assert_called_once_with(xml_file=xml_file, creds=creds, interrupt_after=None, skip_validation=True)


@patch("dsp_tools.cli.call_action.ingest_xmlupload")
def test_ingest_xmlupload_interrupt_after(ingest_xmlupload: Mock) -> None:
    xml_file = Path("filename.xml")
    args = f"ingest-xmlupload --interrupt-after=1 {xml_file}".split()
    entry_point.run(args)
    creds = ServerCredentials(
        server="http://0.0.0.0:3333",
        user="root@example.com",
        password="test",
        dsp_ingest_url="http://0.0.0.0:3340",
    )
    ingest_xmlupload.assert_called_once_with(xml_file=xml_file, creds=creds, interrupt_after=1, skip_validation=False)


@patch("dsp_tools.cli.call_action.ingest_xmlupload")
def test_ingest_xmlupload_remote(ingest_xmlupload: Mock) -> None:
    xml_file = Path("filename.xml")
    server = "https://api.test.dasch.swiss"
    user = "first-name.second-name@dasch.swiss"
    password = "foobar"
    args = f"ingest-xmlupload --server={server} --user {user} --password={password} {xml_file}".split()
    entry_point.run(args)
    creds = ServerCredentials(
        server=server,
        user=user,
        password=password,
        dsp_ingest_url=server.replace("api", "ingest"),
    )
    ingest_xmlupload.assert_called_once_with(
        xml_file=xml_file, creds=creds, interrupt_after=None, skip_validation=False
    )


@patch("dsp_tools.cli.call_action.excel2json")
def test_excel2json(excel2json: Mock) -> None:
    folder = "folder"
    out_file = "filename.json"
    args = f"excel2json {folder} {out_file}".split()
    entry_point.run(args)
    excel2json.assert_called_once_with(
        data_model_files=folder,
        path_to_output_file=out_file,
    )


@patch("dsp_tools.cli.call_action.excel2lists")
def test_excel2lists(excel2lists: Mock) -> None:
    excel2lists.return_value = ([], True)
    file = "filename.xlsx"
    out_file = "filename.json"
    args = f"excel2lists {file} {out_file}".split()
    entry_point.run(args)
    excel2lists.assert_called_once_with(
        excelfolder=file,
        path_to_output_file=out_file,
    )


@patch("dsp_tools.cli.call_action.excel2resources")
def test_excel2resources(excel2resources: Mock) -> None:
    excel2resources.return_value = ([], True)
    file = "filename.xlsx"
    out_file = "filename.json"
    args = f"excel2resources {file} {out_file}".split()
    entry_point.run(args)
    excel2resources.assert_called_once_with(
        excelfile=file,
        path_to_output_file=out_file,
    )


@patch("dsp_tools.cli.call_action.excel2properties")
def test_excel2properties(excel2properties: Mock) -> None:
    excel2properties.return_value = ([], True)
    file = "filename.xlsx"
    out_file = "filename.json"
    args = f"excel2properties {file} {out_file}".split()
    entry_point.run(args)
    excel2properties.assert_called_once_with(
        excelfile=file,
        path_to_output_file=out_file,
    )


@patch("dsp_tools.cli.call_action.old_excel2json")
def test_old_excel2json(old_excel2json: Mock) -> None:
    folder = "folder"
    out_file = "filename.json"
    args = f"old-excel2json {folder} {out_file}".split()
    entry_point.run(args)
    old_excel2json.assert_called_once_with(
        data_model_files=folder,
        path_to_output_file=out_file,
    )


@patch("dsp_tools.cli.call_action.old_excel2lists")
def test_old_excel2lists(old_excel2lists: Mock) -> None:
    old_excel2lists.return_value = ([], True)
    file = "filename.xlsx"
    out_file = "filename.json"
    args = f"old-excel2lists {file} {out_file}".split()
    entry_point.run(args)
    old_excel2lists.assert_called_once_with(
        excelfolder=file,
        path_to_output_file=out_file,
        verbose=False,
    )


@patch("dsp_tools.cli.call_action.id2iri")
def test_id2iri_default(id2iri: Mock) -> None:
    xml_file = "filename.xml"
    json_file = "filename.json"
    args = f"id2iri {xml_file} {json_file}".split()
    entry_point.run(args)
    id2iri.assert_called_once_with(
        xml_file=xml_file,
        json_file=json_file,
        remove_resource_if_id_in_mapping=False,
    )


@patch("dsp_tools.cli.call_action.id2iri")
def test_id2iri_remove_resources(id2iri: Mock) -> None:
    xml_file = "filename.xml"
    json_file = "filename.json"
    args = f"id2iri --remove-resources {xml_file} {json_file}".split()
    entry_point.run(args)
    id2iri.assert_called_once_with(
        xml_file=xml_file,
        json_file=json_file,
        remove_resource_if_id_in_mapping=True,
    )


@patch("dsp_tools.cli.call_action.excel2xml", return_value=("foo", "bar"))
def test_excel2xml(excel2xml: Mock) -> None:
    excel_file = "filename.xlsx"
    shortcode = "1234"
    onto = "someonto"
    args = f"excel2xml {excel_file} {shortcode} {onto}".split()
    entry_point.run(args)
    excel2xml.assert_called_once_with(
        datafile=excel_file,
        shortcode=shortcode,
        default_ontology=onto,
    )


@patch("dsp_tools.commands.start_stack.StackHandler.start_stack")
@patch("dsp_tools.commands.start_stack.StackHandler.__init__", return_value=None)
def test_start_stack_default(mock_init: Mock, start_stack: Mock) -> None:
    args = "start-stack".split()
    entry_point.run(args)
    mock_init.assert_called_once_with(
        StackConfiguration(
            max_file_size=None,
            enforce_docker_system_prune=False,
            suppress_docker_system_prune=False,
            latest_dev_version=False,
            upload_test_data=False,
        )
    )
    start_stack.assert_called_once()


@patch("dsp_tools.commands.start_stack.StackHandler.start_stack")
@patch("dsp_tools.commands.start_stack.StackHandler.__init__", return_value=None)
def test_start_stack_max_file_size(mock_init: Mock, start_stack: Mock) -> None:
    args = "start-stack --max_file_size=1".split()
    entry_point.run(args)
    mock_init.assert_called_once_with(
        StackConfiguration(
            max_file_size=1,
            enforce_docker_system_prune=False,
            suppress_docker_system_prune=False,
            latest_dev_version=False,
            upload_test_data=False,
        )
    )
    start_stack.assert_called_once()


@patch("dsp_tools.commands.start_stack.StackHandler.start_stack")
@patch("dsp_tools.commands.start_stack.StackHandler.__init__", return_value=None)
def test_start_stack_prune(mock_init: Mock, start_stack: Mock) -> None:
    args = "start-stack --prune".split()
    entry_point.run(args)
    mock_init.assert_called_once_with(
        StackConfiguration(
            max_file_size=None,
            enforce_docker_system_prune=True,
            suppress_docker_system_prune=False,
            latest_dev_version=False,
            upload_test_data=False,
        )
    )
    start_stack.assert_called_once()


@patch("dsp_tools.commands.start_stack.StackHandler.start_stack")
@patch("dsp_tools.commands.start_stack.StackHandler.__init__", return_value=None)
def test_start_stack_no_prune(mock_init: Mock, start_stack: Mock) -> None:
    args = "start-stack --no-prune".split()
    entry_point.run(args)
    mock_init.assert_called_once_with(
        StackConfiguration(
            max_file_size=None,
            enforce_docker_system_prune=False,
            suppress_docker_system_prune=True,
            latest_dev_version=False,
            upload_test_data=False,
        )
    )
    start_stack.assert_called_once()


@patch("dsp_tools.commands.start_stack.StackHandler.start_stack")
@patch("dsp_tools.commands.start_stack.StackHandler.__init__", return_value=None)
def test_start_stack_latest(mock_init: Mock, start_stack: Mock) -> None:
    args = "start-stack --latest".split()
    entry_point.run(args)
    mock_init.assert_called_once_with(
        StackConfiguration(
            max_file_size=None,
            enforce_docker_system_prune=False,
            suppress_docker_system_prune=False,
            latest_dev_version=True,
            upload_test_data=False,
        )
    )
    start_stack.assert_called_once()


@patch("dsp_tools.commands.start_stack.StackHandler.start_stack")
@patch("dsp_tools.commands.start_stack.StackHandler.__init__", return_value=None)
def test_start_stack_with_test_data(mock_init: Mock, start_stack: Mock) -> None:
    args = "start-stack --with-test-data".split()
    entry_point.run(args)
    mock_init.assert_called_once_with(
        StackConfiguration(
            max_file_size=None,
            enforce_docker_system_prune=False,
            suppress_docker_system_prune=False,
            latest_dev_version=False,
            upload_test_data=True,
        )
    )
    start_stack.assert_called_once()


@patch("dsp_tools.commands.start_stack.StackHandler.stop_stack")
def test_stop_stack(stop_stack: Mock) -> None:
    args = "stop-stack".split()
    entry_point.run(args)
    stop_stack.assert_called_once_with()


@patch("dsp_tools.cli.call_action.generate_template_repo")
def test_template(generate_template_repo: Mock) -> None:
    args = "template".split()
    entry_point.run(args)
    generate_template_repo.assert_called_once_with()


@patch("dsp_tools.cli.call_action.upload_rosetta")
def test_rosetta(upload_rosetta: Mock) -> None:
    args = "rosetta".split()
    entry_point.run(args)
    upload_rosetta.assert_called_once_with()


@patch("dsp_tools.cli.call_action.xmlupload")
@patch("dsp_tools.cli.entry_point._check_version")
def test_suppress_update_prompt_flag_absent(check_version: Mock, xmlupload: Mock) -> None:
    args = "xmlupload --user=testuser data.xml".split()
    entry_point.run(args)
    check_version.assert_called_once()
    xmlupload.assert_called_once()


@patch("dsp_tools.cli.call_action.xmlupload")
@patch("dsp_tools.cli.entry_point._check_version")
def test_suppress_update_prompt_leftmost(check_version: Mock, xmlupload: Mock) -> None:
    args = "xmlupload --suppress-update-prompt --user=testuser data.xml".split()
    entry_point.run(args)
    check_version.assert_not_called()
    xmlupload.assert_called_once()


@patch("dsp_tools.cli.call_action.xmlupload")
@patch("dsp_tools.cli.entry_point._check_version")
def test_suppress_update_prompt_middle(check_version: Mock, xmlupload: Mock) -> None:
    args = "xmlupload --user=testuser --suppress-update-prompt data.xml".split()
    entry_point.run(args)
    check_version.assert_not_called()
    xmlupload.assert_called_once()


@patch("dsp_tools.cli.call_action.xmlupload")
@patch("dsp_tools.cli.entry_point._check_version")
def test_suppress_update_prompt_rightmost(check_version: Mock, xmlupload: Mock) -> None:
    args = "xmlupload --user=testuser data.xml --suppress-update-prompt".split()
    entry_point.run(args)
    check_version.assert_not_called()
    xmlupload.assert_called_once()


if __name__ == "__main__":
    pytest.main([__file__])



# mypy: disable-error-code="no-untyped-def"

from pathlib import Path

import pytest

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.cli.args import ValidateDataConfig
from dsp_tools.cli.args import ValidationSeverity
from dsp_tools.clients.authentication_client import AuthenticationClient
from dsp_tools.clients.authentication_client_live import AuthenticationClientLive
from dsp_tools.commands.validate_data.shacl_cli_validator import ShaclCliValidator
from dsp_tools.commands.validate_data.validate_data import _validate_data
from test.e2e.commands.validate_data.util import prepare_data_for_validation_from_file

# ruff: noqa: ARG001 Unused function argument

CONFIG = ValidateDataConfig(
    xml_file=Path(),
    save_graph_dir=None,
    severity=ValidationSeverity.INFO,
    ignore_duplicate_files_warning=False,
    is_on_prod_server=True,
)


@pytest.fixture(scope="module")
def authentication(creds: ServerCredentials) -> AuthenticationClient:
    auth = AuthenticationClientLive(server=creds.server, email=creds.user, password=creds.password)
    return auth


@pytest.mark.usefixtures("create_generic_project")
def test_minimal_correct(authentication) -> None:
    file = Path("testdata/validate-data/generic/minimal_correct.xml")
    graphs, used_iris = prepare_data_for_validation_from_file(
        file, authentication, CONFIG.ignore_duplicate_files_warning
    )
    validation_result = _validate_data(graphs, used_iris, CONFIG)
    assert validation_result.no_problems
    assert not validation_result.problems
    assert not validation_result.report_graphs


@pytest.mark.usefixtures("create_generic_project")
def test_cardinality_correct(authentication, shacl_validator: ShaclCliValidator) -> None:
    file = Path("testdata/validate-data/generic/cardinality_correct.xml")
    graphs, used_iris = prepare_data_for_validation_from_file(
        file, authentication, CONFIG.ignore_duplicate_files_warning
    )
    validation_result = _validate_data(graphs, used_iris, CONFIG)
    assert validation_result.no_problems
    assert not validation_result.problems
    assert not validation_result.report_graphs


@pytest.mark.usefixtures("create_generic_project")
def test_content_correct(authentication, shacl_validator: ShaclCliValidator) -> None:
    file = Path("testdata/validate-data/generic/content_correct.xml")
    graphs, used_iris = prepare_data_for_validation_from_file(
        file, authentication, CONFIG.ignore_duplicate_files_warning
    )
    validation_result = _validate_data(graphs, used_iris, CONFIG)
    assert validation_result.no_problems
    assert not validation_result.problems
    assert not validation_result.report_graphs


@pytest.mark.usefixtures("create_generic_project")
def test_file_value_correct(authentication, shacl_validator: ShaclCliValidator) -> None:
    file = Path("testdata/validate-data/generic/file_value_correct.xml")
    graphs, used_iris = prepare_data_for_validation_from_file(
        file, authentication, CONFIG.ignore_duplicate_files_warning
    )
    validation_result = _validate_data(graphs, used_iris, CONFIG)
    assert validation_result.no_problems
    assert not validation_result.problems
    assert not validation_result.report_graphs


@pytest.mark.usefixtures("create_generic_project")
def test_dsp_inbuilt_correct(authentication, shacl_validator: ShaclCliValidator) -> None:
    file = Path("testdata/validate-data/generic/dsp_inbuilt_correct.xml")
    graphs, used_iris = prepare_data_for_validation_from_file(
        file, authentication, CONFIG.ignore_duplicate_files_warning
    )
    validation_result = _validate_data(graphs, used_iris, CONFIG)
    assert validation_result.no_problems
    assert not validation_result.problems
    assert not validation_result.report_graphs



from dataclasses import dataclass

from typing_extensions import deprecated

from dsp_tools.error.xmllib_warnings import MessageInfo
from dsp_tools.error.xmllib_warnings_util import emit_xmllib_input_warning
from dsp_tools.xmllib.value_checkers import is_dsp_ark
from dsp_tools.xmllib.value_checkers import is_dsp_iri
from dsp_tools.xmllib.value_checkers import is_timestamp

LIST_SEPARATOR = "\n    - "


@dataclass
class MigrationMetadata:
    """
    Metadata originating from a SALSAH migration.
    This will be deprecated in the future when all projects are migrated.
    """

    creation_date: str | None
    iri: str | None
    ark: str | None
    res_id: str

    @deprecated("This is for salsah migration only and will be deleted in future releases.")
    def __post_init__(self) -> None:
        msg_list = []
        if self.creation_date and not is_timestamp(self.creation_date):
            msg_list.append(f"The value for creation date is not a valid timestamp: {self.creation_date}")
        if self.iri and not is_dsp_iri(self.iri):
            msg_list.append(f"The provided IRI is not valid: {self.iri}")
        if self.ark and not is_dsp_ark(self.ark):
            msg_list.append(f"The provided ARK is not valid: {self.ark}")
        if msg_list:
            msg = f"The migration metadata has the following problem(s):{LIST_SEPARATOR}{LIST_SEPARATOR.join(msg_list)}"
            emit_xmllib_input_warning(MessageInfo(msg, self.res_id))

    def as_attrib(self) -> dict[str, str]:
        attrib_dict = {}
        if self.creation_date:
            attrib_dict["creation_date"] = self.creation_date
        if self.iri:
            attrib_dict["iri"] = self.iri
        if self.ark:
            attrib_dict["ark"] = self.ark
        if not attrib_dict:
            msg = (
                "The migration metadata of the resource does not contain any values. Please check if an error occurred."
            )
            emit_xmllib_input_warning(MessageInfo(msg, self.res_id))
        return attrib_dict



# mypy: disable-error-code="no-untyped-def"

from collections.abc import Iterator
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest
import requests

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.commands.xmlupload.xmlupload import xmlupload

# ruff: noqa: ARG001 Unused function argument

PROJECT_SHORTCODE = "9999"
ONTO_NAME = "onto"
SECOND_ONTO = "second-onto"


@pytest.fixture(scope="module")
def project_iri(create_generic_project, creds: ServerCredentials) -> str:
    get_project_route = f"{creds.server}/admin/projects/shortcode/{PROJECT_SHORTCODE}"
    project_iri: str = requests.get(get_project_route, timeout=3).json()["project"]["id"]
    return project_iri


@pytest.fixture(scope="module")
def onto_iri(creds) -> str:
    return f"{creds.server}/ontology/{PROJECT_SHORTCODE}/{ONTO_NAME}/v2#"


@pytest.fixture(scope="module")
def class_with_everything_iri(onto_iri) -> str:
    return f"{onto_iri}ClassWithEverything"


@pytest.fixture(scope="module")
def second_onto_iri(creds) -> str:
    return f"{creds.server}/ontology/{PROJECT_SHORTCODE}/{SECOND_ONTO}/v2#"


@pytest.fixture(scope="module")
def _xmlupload_minimal_correct(create_generic_project, creds) -> None:
    """
    If there is more than 1 module, pytest-xdist might execute this fixture for multiple modules at the same time.
    This can lead to the situation that multiple workers start the xmlupload of the same data at the same time.
    Then it can happen that they try to save the id2iri mapping at the same time,
    which fails, because the id2iri mapping is named after the shortcode and the timestamp.
    """
    absolute_xml_path = Path("testdata/validate-data/generic/minimal_correct.xml").absolute()
    original_cwd = Path.cwd()
    with TemporaryDirectory() as tmpdir:
        with pytest.MonkeyPatch.context() as m:
            m.chdir(tmpdir)
            assert xmlupload(absolute_xml_path, creds, str(original_cwd))


@pytest.fixture(scope="module")
def _xmlupload_text_parsing(create_generic_project: None, creds: ServerCredentials) -> Iterator[None]:
    assert xmlupload(Path("testdata/xml-data/generic_project_text_parsing.xml"), creds, ".")
    yield
    if found := list(Path.cwd().glob(f"id2iri_{PROJECT_SHORTCODE}_localhost_*.json")):
        found[0].unlink()


@pytest.fixture(scope="module")
def auth_header(create_generic_project, creds) -> dict[str, str]:
    payload = {"email": creds.user, "password": creds.password}
    token: str = requests.post(f"{creds.server}/v2/authentication", json=payload, timeout=3).json()["token"]
    return {"Authorization": f"Bearer {token}"}



from typing import Any
from unittest.mock import Mock
from unittest.mock import patch

import pytest

from dsp_tools.commands.validate_data.api_clients import ListClient
from dsp_tools.commands.validate_data.api_clients import OntologyClient
from dsp_tools.error.exceptions import InternalError


@pytest.fixture
def api_url() -> str:
    return "http://0.0.0.0:3333"


@pytest.fixture
def ontology_client(api_url: str) -> OntologyClient:
    return OntologyClient(api_url, "9999")


@pytest.fixture
def list_client(api_url: str) -> ListClient:
    return ListClient(api_url, "9999")


class TestOntologyClient:
    def test_get_ontology_iris_ok(self, ontology_client: OntologyClient) -> None:
        mock_response = Mock(status_code=200, ok=True, headers={})
        mock_response.json.return_value = {"project": {"ontologies": ["onto_iri"]}}
        with patch("dsp_tools.commands.validate_data.api_clients.requests.get", return_value=mock_response) as mock_get:
            result = ontology_client._get_ontology_iris()
        assert result == ["onto_iri"]
        assert mock_get.call_args_list[0][1]["url"] == f"{ontology_client.api_url}/admin/projects/shortcode/9999"

    def test_get_ontology_iris_non_ok_code(self, ontology_client: OntologyClient) -> None:
        mock_response = Mock(status_code=404, ok=False, headers={})
        mock_response.json.return_value = {}
        with patch("dsp_tools.commands.validate_data.api_clients.requests.get", return_value=mock_response):
            with pytest.raises(InternalError):
                ontology_client._get_ontology_iris()

    def test_get_ontology_iris_no_ontology_key(self, ontology_client: OntologyClient) -> None:
        mock_response = Mock(status_code=200, ok=True, headers={})
        mock_response.json.return_value = {"foo": "bar"}
        with patch("dsp_tools.commands.validate_data.api_clients.requests.get", return_value=mock_response):
            with pytest.raises(InternalError):
                ontology_client._get_ontology_iris()

    def test_get_one_ontology(self, ontology_client: OntologyClient) -> None:
        mock_response = Mock(status_code=200, ok=True, headers={}, text="Turtle Text")
        mock_response.json.return_value = {}
        with patch("dsp_tools.commands.validate_data.api_clients.requests.get", return_value=mock_response) as mock_get:
            result = ontology_client._get_one_ontology("iri")
        assert result == "Turtle Text"
        assert mock_get.call_args_list[0][1]["url"] == "iri"
        assert mock_get.call_args_list[0][1]["headers"] == {"Accept": "text/turtle"}


class TestListConnection:
    def test_get_all_list_iris(self, list_client: ListClient) -> None:
        mock_response = Mock(status_code=200, ok=True, headers={})
        mock_response.json.return_value = {"lists": []}
        with patch("dsp_tools.commands.validate_data.api_clients.requests.get", return_value=mock_response) as mock_get:
            result = list_client._get_all_list_iris()
        assert result == {"lists": []}
        assert mock_get.call_args_list[0][1]["url"] == f"{list_client.api_url}/admin/lists?projectShortcode=9999"

    def test_get_all_list_iris_non_ok_code(self, list_client: ListClient) -> None:
        mock_response = Mock(status_code=404, ok=False, headers={})
        mock_response.json.return_value = {}
        with patch("dsp_tools.commands.validate_data.api_clients.requests.get", return_value=mock_response):
            with pytest.raises(InternalError):
                list_client._get_all_list_iris()

    def test_get_one_list(self, list_client: ListClient) -> None:
        mock_response = Mock(status_code=200, ok=True, headers={})
        mock_response.json.return_value = {"type": "ListGetResponseADM", "list": {}}
        with patch("dsp_tools.commands.validate_data.api_clients.requests.get", return_value=mock_response) as mock_get:
            result = list_client._get_one_list("http://rdfh.ch/lists/9999/WWqeCEj8R_qrK5djsVcHvg")
        assert result == {"type": "ListGetResponseADM", "list": {}}
        url_expected = f"{list_client.api_url}/admin/lists/http%3A%2F%2Frdfh.ch%2Flists%2F9999%2FWWqeCEj8R_qrK5djsVcHvg"
        assert mock_get.call_args_list[0][1]["url"] == url_expected

    def test_get_one_list_non_ok_code(self, list_client: ListClient) -> None:
        mock_response = Mock(status_code=404, ok=False, headers={})
        mock_response.json.return_value = {}
        with patch("dsp_tools.commands.validate_data.api_clients.requests.get", return_value=mock_response):
            with pytest.raises(InternalError):
                list_client._get_one_list("http://rdfh.ch/lists/9999/WWqeCEj8R_qrK5djsVcHvg")

    def test_extract_list_iris(self, list_client: ListClient, response_all_list_one_project: dict[str, Any]) -> None:
        extracted = list_client._extract_list_iris(response_all_list_one_project)
        expected = {"http://rdfh.ch/lists/9999/list1", "http://rdfh.ch/lists/9999/list2"}
        assert set(extracted) == expected

    def test_extract_list_iris_no_lists(
        self, list_client: ListClient, response_all_list_one_project_no_lists: dict[str, Any]
    ) -> None:
        extracted = list_client._extract_list_iris(response_all_list_one_project_no_lists)
        assert not extracted

    def test_reformat_one_list(self, list_client: ListClient, response_one_list: dict[str, Any]) -> None:
        reformatted = list_client._reformat_one_list(response_one_list)
        sorted_nodes = sorted(reformatted.nodes, key=lambda x: x.name)
        assert reformatted.list_iri == "http://rdfh.ch/lists/9999/list1"
        assert reformatted.list_name == "firstList"
        names = [x.name for x in sorted_nodes]
        assert names == ["n1", "n1.1", "n1.1.1", "n1.1.2"]
        expected_iris = [
            "http://rdfh.ch/lists/9999/n1",
            "http://rdfh.ch/lists/9999/n11",
            "http://rdfh.ch/lists/9999/n111",
            "http://rdfh.ch/lists/9999/n112",
        ]
        iris = [x.iri for x in sorted_nodes]
        assert iris == expected_iris


if __name__ == "__main__":
    pytest.main([__file__])



from dataclasses import dataclass
from pathlib import Path

import pandas as pd

from dsp_tools.error.problems import Problem

separator = "\n\n"
list_separator = "\n - "


@dataclass(frozen=True)
class FileProblems(Problem):
    """Handle the error communication to the user in case that some files don't exist or are unsupported."""

    non_existing_files: list[Path]
    unsupported_files: list[Path]
    maximum_prints: int = 50

    def __post_init__(self) -> None:
        if not self.non_existing_files and not self.unsupported_files:
            raise ValueError("It's not possible to create a FileProblems object without any problems.")

    def execute_error_protocol(self) -> str:
        """
        Generate the error message to communicate the problems to the user.
        If there are too many problems, save them to a file.

        Returns:
            error message
        """
        msg = "Some files referenced in the <bitstream> tags of your XML file cannot be uploaded to the server."
        if len(self.non_existing_files) + len(self.unsupported_files) > self.maximum_prints:
            output_file = Path("file_problems.csv")
            self._save_to_csv(output_file)
            msg += f" The full list of files with problems has been saved to '{output_file}'."
            return msg
        if self.non_existing_files:
            msg += separator
            msg += "The following files don't exist on your computer:"
            msg += list_separator + list_separator.join([str(file) for file in self.non_existing_files])
        if self.unsupported_files:
            msg += separator
            msg += "The following files have unsupported extensions:"
            msg += list_separator + list_separator.join([str(file) for file in self.unsupported_files])
        return msg

    def _save_to_csv(self, output_file: Path) -> None:
        problems = ["File doesn't exist"] * len(self.non_existing_files) + ["Extension not supported"] * len(
            self.unsupported_files
        )
        data = {
            "File": self.non_existing_files + self.unsupported_files,
            "Problem": problems,
        }
        df = pd.DataFrame(data)
        df.to_csv(output_file, index=False)



from collections.abc import Iterator
from pathlib import Path

import pytest

from dsp_tools.xmllib.models.res import Resource
from dsp_tools.xmllib.models.root import XMLRoot


@pytest.fixture
def out_file() -> Iterator[Path]:
    out_file = Path("testdata/xml-data/tmp-serialise_tags.xml")
    yield out_file
    out_file.unlink()


@pytest.mark.parametrize(
    ("richtext", "expected"),
    [
        (
            'Text <footnote content="Text &lt;a href=&quot;https://www.google.com/&quot;&gt;Google&lt;/a&gt;" />',
            'Text <footnote content="Text &lt;a href=&quot;https://www.google.com/&quot;&gt;Google&lt;/a&gt;"/>',
            # &quot; is always resolved by the XML parser, but then automatically re-escaped
            # because it is part of an attribute
        ),
        (
            'Text <footnote content="Text &lt;a href=&#34;https://www.google.com/&#34;&gt;Google&lt;/a&gt;" />',
            'Text <footnote content="Text &lt;a href=&quot;https://www.google.com/&quot;&gt;Google&lt;/a&gt;"/>',
            # &#34; is always resolved by the XML parser, but then automatically re-escaped to &quot;
            # because it is part of an attribute
        ),
        (
            "Text &quot; &lt;",
            'Text " &lt;',
            # &quot; is always resolved by the XML parser, if it is part of a text
        ),
        (
            'Text <a href="https://www.google.com/">Google</a>',
            'Text <a href="https://www.google.com/">Google</a>',
        ),
        (
            "Text <p><strong><em>Text</em></strong></p>",
            "Text <p><strong><em>Text</em></strong></p>",
        ),
    ],
)
def test_serialisation_of_tags(richtext: str, expected: str, out_file: Path) -> None:
    root = XMLRoot.create_new("1234", "my_onto")
    res = Resource.create_new("res_id", ":Type", "lbl").add_richtext(":prop", richtext)
    root.add_resource(res).write_file(out_file)
    with open(out_file, encoding="utf-8") as f:
        lines = f.readlines()
    start_index = next(i for i, x in enumerate(lines) if '<text encoding="xml">' in x)
    end_index = next(i for i, x in enumerate(lines) if "</text>" in x)
    lines_purged = [x.strip() for x in lines[start_index : end_index + 1]]
    result = "".join(lines_purged).replace('<text encoding="xml">', "").replace("</text>", "")
    assert result == expected



from typing import Any

import pandas as pd
import regex

from dsp_tools.error.xmllib_warnings import MessageInfo
from dsp_tools.error.xmllib_warnings_util import emit_xmllib_input_warning
from dsp_tools.xmllib.internal.circumvent_circular_imports import parse_richtext_as_xml


def is_nonempty_value(value: Any) -> bool:
    """
    Check if a value is not None-like
    or that its string representation contains at least one of the following characters:

    - ``\\p{S}`` = symbols and special characters
    - ``\\p{P}`` = punctuation
    - ``\\w`` = all Unicode letters, numbers, and _

    Args:
        value: value of any type

    Returns:
        True if the value is not None-like and contains at least one of the above-mentioned characters

    Examples:
        ```python
        # True values:
        assert xmllib.is_nonempty_value("word") == True
        assert xmllib.is_nonempty_value("None") == True
        assert xmllib.is_nonempty_value("-") == True
        assert xmllib.is_nonempty_value(0) == True
        assert xmllib.is_nonempty_value(1) == True
        assert xmllib.is_nonempty_value("0") == True
        assert xmllib.is_nonempty_value("1") == True
        assert xmllib.is_nonempty_value(True) == True
        assert xmllib.is_nonempty_value(False) == True
        assert xmllib.is_nonempty_value("עִבְרִית") == True

        # False values:
        assert xmllib.is_nonempty_value(pd.NA) == False
        assert xmllib.is_nonempty_value(None) == False
        assert xmllib.is_nonempty_value("") == False
        assert xmllib.is_nonempty_value(" ") == False
        assert xmllib.is_nonempty_value("\\n") == False
        ```
    """
    if pd.isna(value):
        return False
    if regex.search(r"[\p{S}\p{P}\w]", str(value), flags=regex.UNICODE):
        return True
    return False


def is_bool_like(value: Any) -> bool:
    """
    Checks if a value is a bool or can be converted into a bool.
    It is case-insensitive, meaning that the words can also be capitalised.

    Accepted values:
         - `false`, `0`, `0.0`, `no`, `non`, `nein` -> `False`
         - `true`, `1`, `1.0`, `yes`, `oui`, `ja` -> `True`

    Args:
        value: value to check

    Returns:
        True if it conforms

    Examples:
        ```python
        result = xmllib.is_bool_like("yes")
        # result == True
        ```

        ```python
        result = xmllib.is_bool_like("not like a bool")
        # result == False
        ```
    """
    value = str(value).lower().strip()
    if value in ("false", "0", "0.0", "no", "non", "nein"):
        return True
    elif value in ("true", "1", "1.0", "yes", "oui", "ja"):
        return True
    return False


def is_color(value: Any) -> bool:
    """
    Checks if a value is a color value.

    Args:
        value: value to check

    Returns:
        True if it conforms

    Examples:
        ```python
        result = xmllib.is_color("#00ff66")
        # result == True
        ```

        ```python
        result = xmllib.is_color("not a color")
        # result == False
        ```
    """
    return bool(regex.search(r"^#[0-9a-f]{6}$", str(value).strip(), flags=regex.IGNORECASE))


def is_date(value: Any) -> bool:
    """
    Checks if a value is a date value.

    Args:
        value: value to check

    Returns:
        True if it conforms

    Examples:
        ```python
        result = xmllib.is_date("GREGORIAN:CE:2014-01-31:CE:2014-01-31")
        # result == True
        ```

        ```python
        result = xmllib.is_date("not a date")
        # result == False
        ```
    """
    calendar_optional = r"(?:(GREGORIAN|JULIAN|ISLAMIC):)?"
    first_era_optional = r"(?:(CE|BCE|BC|AD):)?"
    second_area_optional = r"(?::(CE|BCE|BC|AD))?"
    date = r"\d{1,4}(?:-\d{1,2}){0,2}"
    date_mandatory = rf"({date})"
    date_optional = rf"(:{date})?"
    full_date_pattern = (
        rf"^{calendar_optional}{first_era_optional}{date_mandatory}{second_area_optional}{date_optional}$"
    )
    found = regex.search(full_date_pattern, str(value))
    if not found:
        return False
    if found.group(1) == "ISLAMIC" and (found.group(2) or found.group(4)):
        # eras are not supported yet for the islamic calendar
        return False
    return True


def is_geoname(value: Any) -> bool:
    """
    Checks if a value is a geoname value.

    Args:
        value: value to check

    Returns:
        True if it conforms

    Examples:
        ```python
        result = xmllib.is_geoname("8879000")
        # result == True
        ```

        ```python
        result = xmllib.is_geoname("not a geoname code")
        # result == False
        ```
    """
    return is_integer(value)


def is_decimal(value: Any) -> bool:
    """
    Checks if a value is a float, an integer, or a string which can be converted into a float.

    Args:
        value: value to check

    Returns:
        True if conforms to the above-mentioned criteria.

    Examples:
        ```python
        result = xmllib.is_decimal("0.1")
        # result == True
        ```

        ```python
        # because this is equivalent to 9.0 it is accepted

        result = xmllib.is_decimal(9)
        # result == True
        ```

        ```python
        result = xmllib.is_decimal("not a decimal")
        # result == False
        ```
    """
    if pd.isna(value):
        return False

    match value:
        case bool():
            return False
        case int() | float():
            return True
    try:
        float(value)
        return True
    except ValueError:
        return False


def is_integer(value: Any) -> bool:
    """
    Checks if a value is an integer or a string which can be converted into an integer.

    Args:
        value: value to check

    Returns:
        True if conforms to the above-mentioned criteria.

    Examples:
        ```python
        result = xmllib.is_integer("1")
        # result == True
        ```

        ```python
        result = xmllib.is_integer(9.1)
        # result == False
        ```

        ```python
        result = xmllib.is_integer("not an integer")
        # result == False
        ```
    """
    match value:
        case bool():
            return False
        case int():
            return True
        case str():
            return bool(regex.search(r"^\d+$", value))
        case _:
            return False


def is_link_value(value: Any) -> bool:
    """
    Check if a value is a valid internal ID of a resource (xsd:ID) or a valid internal DSP IRI.
    Both of these values are allowed in LinkValues.

    Args:
        value: the target ID of a LinkValue

    Returns:
        True if it is a permissible value.

    Examples:
        ```python
        result = xmllib.is_link_value("1_must_not_start_with_number")
        # result == False
        ```

        ```python
        result = xmllib.is_link_value("characters|not|allowed")
        # result == False
        ```

        ```python
        result = xmllib.is_link_value("resource_id_1")
        # result == True
        ```

        ```python
        result = xmllib.is_link_value("http://rdfh.ch/4123/54SYvWF0QUW6a")
        # result == True
        ```
    """
    if is_valid_resource_id(value):
        return True
    return is_dsp_iri(value)


def is_valid_resource_id(value: Any) -> bool:
    """
    Check if a value is a valid internal ID of a resource (xsd:ID).

    Args:
        value: the ID of a Resource

    Returns:
        True if it is a permissible value.

    Examples:
        ```python
        result = xmllib.is_valid_resource_id("1_must_not_start_with_number")
        # result == False
        ```

        ```python
        result = xmllib.is_valid_resource_id("characters|not|allowed")
        # result == False
        ```

        ```python
        result = xmllib.is_valid_resource_id("resource_id_1")
        # result == True
        ```
    """
    if is_nonempty_value(value):
        # None, etc. would not be recognised as invalid since it is converted into a string.
        return bool(
            regex.search(r"^[a-zA-Zçéàèöäüòôûâêñ_][a-zA-Zçéàèöäüòôûâêñ_\d.\-]*$", str(value)),
        )
    return False


def is_timestamp(value: Any) -> bool:
    """
    Checks if a value is a valid timestamp.

    Args:
        value: value to check

    Returns:
        True if it conforms

    Examples:
        ```python
        result = xmllib.is_timestamp("2019-10-23T13:45:12Z")
        # result == True
        ```

        ```python
        result = xmllib.is_timestamp("not a time stamp")
        # result == False
        ```
    """
    validation_regex = r"^\d{4}-[0-1]\d-[0-3]\dT[0-2]\d:[0-5]\d:[0-5]\d(\.\d{1,12})?(Z|[+-][0-1]\d:[0-5]\d)$"
    return bool(regex.search(validation_regex, str(value)))


def is_dsp_iri(value: Any) -> bool:
    """
    Check if a value is a valid internal DSP IRI.

    Args:
        value: IRI

    Returns:
        True if it is valid, else false

    Examples:
        ```python
        result = xmllib.is_dsp_iri("http://rdfh.ch/4123/54SYvWF0QUW6a")
        # result == True
        ```

        ```python
        result = xmllib.is_dsp_iri("http://dbpedia.org/resource/Internationalized_Resource_Identifier")
        # result == False
        ```
    """
    return bool(regex.search(r"^http://rdfh\.ch/[\dA-F]{4}/", str(value)))


def is_dsp_ark(value: Any) -> bool:
    """
    Checks if a value is a valid ARK.

    Args:
        value: ARK

    Returns:
        True if it is valid, else false

    Examples:
        ```python
        result = xmllib.is_dsp_ark("ark:/72163/4123-31ec6eab334-a.2022829")
        # result == True
        ```

        ```python
        result = xmllib.is_dsp_ark("http://rdfh.ch/4123/54SYvWF0QUW6a")
        # result == False
        ```
    """
    return bool(regex.search(r"^ark:/", str(value)))


def check_richtext_syntax(richtext: str) -> None:
    """
    DSP richtexts must be convertible into valid XML.
    This checker escapes the reserved characters `<`, `>` and `&`,
    but only if they are not part of a standard standoff tag or escape sequence.
    Then, it tries to parse the resulting XML.

    Note: Only DSP standard standoff tags are allowed in richtexts. They are documented
    [here](https://docs.dasch.swiss/latest/DSP-API/03-endpoints/api-v2/text/standard-standoff/).

    Args:
        richtext: richtext to check

    Warns:
        XmllibInputWarning: if the input contains XML syntax problems
    """
    result = parse_richtext_as_xml(richtext)
    if isinstance(result, MessageInfo):
        emit_xmllib_input_warning(result)



# mypy: disable-error-code="no-untyped-def"

import json
import urllib.parse

import requests
from rdflib import RDFS
from rdflib import Graph
from rdflib import Literal

from dsp_tools.utils.rdflib_constants import KNORA_API
from dsp_tools.utils.rdflib_constants import SubjectObjectTypeAlias


def util_request_resources_by_class(resclass_iri: str, auth_header: dict[str, str], project_iri: str, creds) -> Graph:
    all_resources = Graph()
    for i in range(100):
        g = _get_single_graph_from_api(resclass_iri, auth_header, project_iri, creds, i)
        all_resources += g
        if not list(g.triples((None, KNORA_API.mayHaveMoreResults, None))):
            break
    return all_resources


def _get_single_graph_from_api(resclass_iri: str, auth_header: dict[str, str], project_iri: str, creds, offset: int):
    resclass_iri_encoded = urllib.parse.quote_plus(resclass_iri)
    get_resources_route = f"{creds.server}/v2/resources?resourceClass={resclass_iri_encoded}&page={offset}"
    headers = auth_header | {"X-Knora-Accept-Project": project_iri}
    response = requests.get(get_resources_route, timeout=3, headers=headers).json()
    resources = json.dumps(response)
    returned_g = Graph()
    returned_g.parse(data=resources, format="json-ld")
    return returned_g


def util_get_res_iri_from_label(g: Graph, label_str: str) -> SubjectObjectTypeAlias:
    return next(g.subjects(RDFS.label, Literal(label_str)))



# mypy: disable-error-code="no-untyped-def"

from pathlib import Path

import pytest

from dsp_tools.cli.args import ValidateDataConfig
from dsp_tools.cli.args import ValidationSeverity
from dsp_tools.commands.validate_data.constants import MAXIMUM_DUPLICATE_FILE_PATHS
from dsp_tools.commands.validate_data.validation.check_duplicate_files import _get_filepaths_with_more_than_one_usage
from dsp_tools.commands.validate_data.validation.check_duplicate_files import check_for_duplicate_files
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedFileValue
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedFileValueMetadata
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedResource

TEST_ENV_CONFIG = ValidateDataConfig(
    xml_file=Path(),
    save_graph_dir=None,
    severity=ValidationSeverity.INFO,
    ignore_duplicate_files_warning=False,
    is_on_prod_server=False,
)

PROD_ENV_CONFIG = ValidateDataConfig(
    xml_file=Path(),
    save_graph_dir=None,
    severity=ValidationSeverity.INFO,
    ignore_duplicate_files_warning=False,
    is_on_prod_server=True,
)

FILEPATH_1 = "file_path_1.jpg"
FILEPATH_2 = "file_path_2.jpg"

TOO_MAY_DUPLICATES_NUMBER = MAXIMUM_DUPLICATE_FILE_PATHS + 1
OK_NUMBER_OF_DUPLICATES = MAXIMUM_DUPLICATE_FILE_PATHS - 1


@pytest.fixture
def file_value_1():
    return ParsedFileValue(FILEPATH_1, None, ParsedFileValueMetadata(None, None, None, None))


@pytest.fixture
def file_value_2():
    return ParsedFileValue(FILEPATH_2, None, ParsedFileValueMetadata(None, None, None, None))


@pytest.fixture
def too_many_duplicate_files(file_value_1):
    return [
        ParsedResource(f"id_{i}", ":type", "lbl", None, [], file_value_1, None)
        for i in range(TOO_MAY_DUPLICATES_NUMBER)
    ] + [ParsedResource(f"no_file_{i}", ":type", "lbl", None, [], None, None) for i in range(10)]


@pytest.fixture
def ok_amount_of_duplicate_files(file_value_1):
    return [
        ParsedResource(f"id_{i}", ":type", "lbl", None, [], file_value_1, None) for i in range(OK_NUMBER_OF_DUPLICATES)
    ]


class TestCheckForDuplicateFiles:
    def test_too_many_files_on_test_environment(self, too_many_duplicate_files):
        result = check_for_duplicate_files(too_many_duplicate_files, TEST_ENV_CONFIG)
        expected_msg = (
            "1 file(s) were used multiple times in your data. "
            "Due to the large number of duplicates they cannot be included in the schema validation.\n"
            "Since you are on a test environment, the validation or xmlupload will continue "
            "without the duplicate check. Please note that this is not allowed on a production server.\n"
            "The following filepaths are used more than once, result displayed as: "
            f"file count - 'file path'\n{TOO_MAY_DUPLICATES_NUMBER} - 'file_path_1.jpg'"
        )
        assert result.user_msg == expected_msg
        assert result.should_continue
        assert result.duplicate_files_must_be_ignored

    def test_too_many_files_on_prod_environment(self, too_many_duplicate_files):
        result = check_for_duplicate_files(too_many_duplicate_files, PROD_ENV_CONFIG)
        expected_msg = (
            "1 file(s) were used multiple times in your data. Due to the large number of "
            "duplicates they cannot be included in the schema validation.\n"
            "Since you are on a production server, the validation or xmlupload cannot continue. "
            "If you wish to upload duplicate images, use the designated flag to ignore them.\n"
            "The following filepaths are used more than once, result displayed as: file count - 'file path'\n"
            f"{TOO_MAY_DUPLICATES_NUMBER} - 'file_path_1.jpg'"
        )
        assert result.user_msg == expected_msg
        assert not result.should_continue
        assert result.duplicate_files_must_be_ignored

    def test_ok_amount_of_duplicate_files_on_test_environment(self, ok_amount_of_duplicate_files):
        result = check_for_duplicate_files(ok_amount_of_duplicate_files, TEST_ENV_CONFIG)
        assert result.user_msg is None
        assert result.should_continue
        assert not result.duplicate_files_must_be_ignored

    def test_ok_amount_of_duplicate_files_on_prod_environment(self, ok_amount_of_duplicate_files):
        result = check_for_duplicate_files(ok_amount_of_duplicate_files, PROD_ENV_CONFIG)
        assert result.user_msg is None
        assert result.should_continue
        assert not result.duplicate_files_must_be_ignored


class TestGetFilePathCountDict:
    def test_several_duplicates(self, file_value_1, file_value_2):
        file_1_number = 10
        file_2_number = 2
        resources = [
            ParsedResource(f"id_{i}", ":type", "lbl", None, [], file_value_1, None) for i in range(file_1_number)
        ]
        resources.extend(
            [ParsedResource(f"id_{i}", ":type", "lbl", None, [], file_value_2, None) for i in range(file_2_number)]
        )
        result = _get_filepaths_with_more_than_one_usage(resources)
        assert set(result.keys()) == {FILEPATH_1, FILEPATH_2}
        assert result[FILEPATH_1] == file_1_number
        assert result[FILEPATH_2] == file_2_number

    def test_one_duplicate(self, file_value_1, file_value_2):
        file_1_number = 10
        file_2_number = 1
        resources = [
            ParsedResource(f"id_{i}", ":type", "lbl", None, [], file_value_1, None) for i in range(file_1_number)
        ]
        resources.extend(
            [ParsedResource(f"id_{i}", ":type", "lbl", None, [], file_value_2, None) for i in range(file_2_number)]
        )
        result = _get_filepaths_with_more_than_one_usage(resources)
        assert set(result.keys()) == {FILEPATH_1}
        assert result[FILEPATH_1] == file_1_number

    def test_no_duplicate(self, file_value_1, file_value_2):
        resources = [
            ParsedResource("id_1", ":type", "lbl", None, [], file_value_1, None),
            ParsedResource("id_2", ":type", "lbl", None, [], file_value_2, None),
        ]
        result = _get_filepaths_with_more_than_one_usage(resources)
        assert not result
